@@ -21,7 +21,7 @@ use List::Util qw(reduce min max pairmap sum);
 use Math::GSL::Matrix;
 use Graphics::GnuplotIF;
 
-our $VERSION = '3.40';
+our $VERSION = '3.41';
 
 @Algorithm::RegressionTree::ISA = ('Algorithm::DecisionTree');
 
@@ -36,6 +36,7 @@ sub new {
     delete $dtargs{predictor_columns};
     delete $dtargs{mse_threshold};
     delete $dtargs{need_data_normalization};
+    delete $dtargs{jacobian_choice};
     delete $dtargs{debug1_r};
     delete $dtargs{debug2_r};
     delete $dtargs{debug3_r};
@@ -44,6 +45,7 @@ sub new {
     $instance->{_dependent_variable_column}       =  $args{dependent_variable_column} || undef;
     $instance->{_predictor_columns}               =  $args{predictor_columns} || 0;
     $instance->{_mse_threshold}                   =  $args{mse_threshold} || 0.01;
+    $instance->{_jacobian_choice}                 =  $args{jacobian_choice} || 0;
     $instance->{_need_data_normalization}         =  $args{need_data_normalization} || 0;
     $instance->{_dependent_var}                   =  undef;
     $instance->{_dependent_var_values}            =  undef;
@@ -58,7 +60,6 @@ sub new {
     bless $instance, $class;
 }
 
-
 ##############################################  Methods  #################################################
 sub get_training_data_for_regression {
     my $self = shift;
@@ -151,7 +152,6 @@ sub get_training_data_for_regression {
     $self->SUPER::calculate_first_order_probabilities();
 }
 
-
 sub construct_XMatrix_and_YVector_all_data {
     my $self = shift;
     my $matrix_rows_as_lists =  [ map {my @arr = @$_; [map substr($_,index($_,'=')+1), @arr] } map {$self->{_training_data_hash}->{$_}} sort {sample_index($a) <=> sample_index($b)} keys %{$self->{_training_data_hash}} ];
@@ -188,6 +188,7 @@ sub estimate_regression_coefficients {
     $display = 0 unless defined $display;
     my ($nrows, $ncols) = $XMatrix->dim;
     print "nrows=$nrows   ncols=$ncols\n" if $self->{_debug2_r};
+    my $jacobian_choice = $self->{_jacobian_choice};
     my $X = $XMatrix->copy();
     my $y = $YVector->copy();
     if ($self->{_need_data_normalization}) {
@@ -196,6 +197,30 @@ sub estimate_regression_coefficients {
     my $beta0 = (transpose($X) * $X)->inverse() * transpose($X) * $y;
     my ($betarows, $betacols) = $beta0->dim;
     die "Something has gone wrong with the calculation of beta coefficients" if $betacols > 1;
+    if ($jacobian_choice == 0) {
+#        my $error = sum(abs_vector_as_list( $y - ($X * $beta) )) / $nrows;   
+        my $error = sum( map abs, ($y - ($X * $beta0) )->col(0)->as_list ) / $nrows;   
+        my $predictions = $X * $beta0;
+        if ($display) {
+            if ($ncols == 2) {
+                my @xvalues = $X->col(0)->as_list;
+                my @yvalues = $predictions->col(0)->as_list;
+                $self->{_output_for_plots}->{scalar(keys %{$self->{_output_for_plots}}) + 1} = [\@xvalues,\@yvalues];
+            } elsif ($ncols == 3) {
+                my @xvalues;
+                my @yvalues = $predictions->col(0)->as_list;
+                foreach my $row_index (0 .. $X->rows - 1) {
+                    my @onerow = $X->row($row_index)->as_list;
+                    pop @onerow;
+                    push @xvalues, "@onerow";
+                }
+                $self->{_output_for_surface_plots}->{scalar(keys %{$self->{_output_for_surface_plots}}) + 1} = [\@xvalues,\@yvalues];
+            } else {
+                print "no display when the overall dimensionality of the data exceeds 3\n";
+            }
+        }
+        return ($error, $beta0);    
+    }
     my $beta = $beta0; 
     if ($self->{_debug2_r}) {
         print "\ndisplaying beta0 matrix\n";
@@ -203,20 +228,39 @@ sub estimate_regression_coefficients {
     }
     my $gamma = 0.1;
     my $iterate_again_flag = 1;
+    my $delta = 0.001;
+    my $master_interation_index = 0;
+    $|++;
     while (1) {
+        print "*" unless $master_interation_index++ % 100;
         last unless $iterate_again_flag;
         $gamma *= 0.1;
         $beta0 = 0.99 * $beta0;
         print "\n\n======== starting iterations with gamma= $gamma ===========\n\n\n" if $self->{_debug2_r};
         $beta = $beta0;
         my $beta_old = Math::GSL::Matrix->new($betarows, 1)->zero;
-        my $error_old = vector_norm( $y - ($X * $beta) );
+        my $error_old = sum( map abs, ($y - ($X * $beta_old) )->col(0)->as_list ) / $nrows;   
         my $error;
         foreach my $iteration (0 .. 1499) {
+            print "." unless $iteration % 100;
             $beta_old = $beta->copy;
-            $beta = $beta_old + 2 * $gamma * transpose($X) * ( $y - ($X * $beta) );
-            $error = vector_norm( $y - ($X * $beta) );
-
+            my $jacobian;
+            if ($jacobian_choice == 1) {
+                $jacobian = $X;
+            } elsif ($jacobian_choice == 2) {      
+                my $x_times_delta_beta = $delta * $X * $beta;
+                $jacobian = Math::GSL::Matrix->new($nrows, $ncols);
+                foreach my $i (0 .. $nrows - 1) {
+                    my @row = ($x_times_delta_beta->get_elem($i,0)) x $ncols;
+                    $jacobian->set_row($i, \@row);
+                }
+                $jacobian = (1.0/$delta) * $jacobian;
+            } else {
+                die "wrong choice for the jacobian_choice";
+            }
+#            $beta = $beta_old + 2 * $gamma * transpose($X) * ( $y - ($X * $beta) );
+            $beta = $beta_old + 2 * $gamma * transpose($jacobian) * ( $y - ($X * $beta) );
+            $error = sum( map abs, ($y - ($X * $beta) )->col(0)->as_list ) / $nrows;   
             if ($error > $error_old) {
                 if (vector_norm($beta - $beta_old) < (0.00001 * vector_norm($beta_old))) {
                     $iterate_again_flag = 0;
@@ -241,7 +285,7 @@ sub estimate_regression_coefficients {
     display_matrix($beta) if $self->{_debug2_r};
     my $predictions = $X * $beta;
     my @error_distribution = ($y - ($X * $beta))->as_list;
-    my $squared_error =  sum map {$_ ** 2} @error_distribution;
+    my $squared_error =  sum map abs, @error_distribution;
     my $error = $squared_error / $nrows;
     if ($display) {
         if ($ncols == 2) {
@@ -256,7 +300,6 @@ sub estimate_regression_coefficients {
                 pop @onerow;
                 push @xvalues, "@onerow";
             }
-#            $self->{_output_for_surface_plots}->{scalar(keys %{$self->{_output_for_surface_plots}}) + 1} = [\@xvalues,\@yvalues];
             $self->{_output_for_surface_plots}->{scalar(keys %{$self->{_output_for_surface_plots}}) + 1} = [\@xvalues,\@yvalues];
         } else {
             print "no display when the overall dimensionality of the data exceeds 3\n";
@@ -799,7 +842,7 @@ sub mse_for_tree_regression_for_all_training_samples {
         $dependent_var_values{"@features_and_vals"} = $dep_value;
         my $leafnode_for_sample = $answer->{'solution_path'}->[-1];
         $leafnode_for_values{"@features_and_vals"} = $answer->{'solution_path'}->[-1];
-        my $error_for_sample = abs($predicted_values{"@features_and_vals"} - $dependent_var_values{"@features_and_vals"}) ** 2;
+        my $error_for_sample = abs($predicted_values{"@features_and_vals"} - $dependent_var_values{"@features_and_vals"});
         $total_error += $error_for_sample;
         if (exists $samples_at_leafnode{$leafnode_for_sample}) {
             push @{$samples_at_leafnode{$leafnode_for_sample}}, $sample;
@@ -811,15 +854,15 @@ sub mse_for_tree_regression_for_all_training_samples {
     my $errors_at_leafnode = { map {$_ => 0.0} @leafnodes_used };
     foreach my $kee (keys %predicted_values) {
         foreach my $leaf (@leafnodes_used) {
-            $errors_at_leafnode->{$leaf} += abs($predicted_values{$kee} - $dependent_var_values{$kee})** 2
+            $errors_at_leafnode->{$leaf} += abs($predicted_values{$kee} - $dependent_var_values{$kee})
                           if $leaf == $leafnode_for_values{$kee};
         }
     }
     my $total_error_per_data_point = $total_error / (scalar keys %{$self->{_training_data_hash}});
-    print "\n\nTotal MSE per sample with tree regression: $total_error_per_data_point\n";
+    print "\n\nTree Regression: Total MSE per sample with tree regression: $total_error_per_data_point\n";
     foreach my $leafnode (@leafnodes_used) {
         my $error_per_data_point = $errors_at_leafnode->{$leafnode} / @{$samples_at_leafnode{$leafnode}};
-        print "MSE per sample at leafnode $leafnode: $error_per_data_point\n";
+        print "    MSE per sample at leafnode $leafnode: $error_per_data_point\n";
     }
     my $error_with_linear_regression = $self->{_root_node}->get_node_error();
     print "For comparision, the MSE per sample error with Linear Regression: $error_with_linear_regression\n";
@@ -1050,6 +1093,7 @@ sub check_for_illegal_params {
                             predictor_columns
                             mse_threshold
                             need_data_normalization
+                            jacobian_choice
                             debug1_r
                             debug2_r
                             debug3_r
@@ -1126,11 +1170,6 @@ sub vector_norm {
     my ($rows, $cols) = $vec->dim;
     die "vector_norm() can only be called for a single column matrix" if $cols > 1;
     my @norm = (transpose($vec) * $vec)->as_list;
-
-#    print "displaying the result of vector norm calculation as list: @norm\n";
-#    my $tempxxx = sqrt($norm[0]);
-#    print "                       will return: $tempxxx\n";
-
     return sqrt($norm[0]);
 }
 
@@ -1142,7 +1181,6 @@ sub display_matrix {
     foreach my $i (0..$nrows-1) {
         my $row = $matrix->row($i);
         my @row_as_list = $row->as_list;
-#        print "@row_as_list\n";
         map { printf("%.4f ", $_) } @row_as_list;
         print "\n";
     }
